% normalization.pl
% strictly after Definition 2.1.18; P. Graf Diss
% Author: pb

:- module(normalization,[
	normalization/4,
	normalization/5
]).


% normalization(Term,NormalTerm,StarSubst,Star0,Star1).
% normalization/4 plus StarSubst, which reconstructs original Term
% from NormalTerm
normalization(Term,NormalTerm,StarSubst,Star0,Star1) :-
	normalization(Term,NormalTerm,[],VarSubst,Star0,Star1),
	reverse_var_subst(VarSubst,StarSubst).

reverse_var_subst([],[]).
reverse_var_subst([X-Y|VarSubst],[Y-X|StarSubst]) :-
	reverse_var_subst(VarSubst,StarSubst).


% normalization(Term,NormalTerm,Star0,Star1)
% replaces each occurrence of each variable in Term
% by a new variable from Star0. If none are left in Star0, 
% a new one is created and appended yielding Star1.
% The leftmost variable is replaced by the first in Star0 and so on.

normalization(Term,NormalTerm,Star0,Star1) :-
	normalization(Term,NormalTerm,[],_VarSubst,Star0,Star1).

normalization(Term,NormalTerm,VarSubst0,VarSubst1,Star0,Star1) :-
	var(Term),!,
	varassoc(VarSubst0,VarSubst1,Term,NormalTerm,Star0,Star1).
normalization(Term,NormalTerm,VarSubst0,VarSubst1,Star0,Star1) :-
	atomic(Term),!,
	Term = NormalTerm,
	VarSubst0 = VarSubst1,
	Star0 = Star1.
normalization(Term,NormalTerm,VarSubst0,VarSubst1,Star0,Star1) :-
	functor(Term,F,A),
	functor(NormalTerm,F,A),
	normalization(1,A,Term,NormalTerm,VarSubst0,VarSubst1,Star0,Star1).

normalization(A1,A,_Term,_NormalTerm,VarSubst0,VarSubst1,Star0,Star1) :- 
	A1 =:= A+1,!,
	VarSubst0 = VarSubst1,
	Star0 = Star1.
normalization(A0,A,Term,NormalTerm,VarSubst0,VarSubst2,Star0,Star2) :-
	arg(A0,Term,SubTerm),
	arg(A0,NormalTerm,SubNormalTerm),
	normalization(SubTerm,SubNormalTerm,VarSubst0,VarSubst1,Star0,Star1),
	A1 is A0 + 1,
	normalization(A1,A,Term,NormalTerm,VarSubst1,VarSubst2,Star1,Star2).
	
	

	

varassoc(V0,V1,T0,T1,Star0,Star1) :-
	varassoc(V0,V1,T0,T1),!, % found an entry
	Star0 = Star1.
varassoc(V0,[V-W|V0],V,W,Star0,Star1) :-
	get_new_star_var(V0,W,Star0,Star1).

% varassoc(V0,V1,T0,T1).
% takes a list of V-W pairs V0 and searches for an
% entry where V = T0. If yes, then W = T1 and V0 = V1, 
% else fails.
varassoc([V-W|V0],[V-W|V1],T0,T1) :-
	(V == T0 ->
	    V0 = V1,
	    T1 = W
	;
	    varassoc(V0,V1,T0,T1)
	).

get_new_star_var([],W,Star0,[W|Star1]) :-
	(Star0 = [W|Star1] -> % there is a star_var left
	    true
	;
	    % Star0 == []; hence W is new star_var
	    Star1 = []
	).
get_new_star_var([_|V0],W,[S0|Star0],[S0|Star1]) :-
	get_new_star_var(V0,W,Star0,Star1).


% Examples
% normalization(h(X,X,Y),NT,[],V,[],Star1).

