:- use_module(substitutiontree).
:- use_module(library(lists)).
:- use_module(sound_unify).

insert_all(Terms,ST) :-
	st_empty_tree(ST0),
	insert_all(Terms,ST0,ST).

insert_all([],ST,ST).
insert_all([Term|Terms],ST0,ST2) :-
	st_insert_term(Term,ST0,ST1),
	write('.'),ttyflush,
	insert_all(Terms,ST1,ST2).


insert_all_info(Terms,ST) :-
	st_empty_tree(ST0),
	insert_all_info(Terms,ST0,ST).

insert_all_info([],ST,ST).
insert_all_info([Term-Info|Terms],ST0,ST2) :-
	st_insert_term(Term,Info,ST0,ST1),
	write('.'),ttyflush,
	insert_all_info(Terms,ST1,ST2).

time_file(File) :-
	cputime(Read0),
	read_terms_file(File,Terms),
	cputime(Read1),
	Read2 is Read1 - Read0,
	format("Time used for file ~w : reading     : ~f seconds~n",[File,Read2]),
	cputime(T),
	insert_all(Terms,ST),
	cputime(TT),
	Time is TT - T,
	format("Time used for file ~w : inserting   : ~f seconds~n",[File,Time]),
%	cputime(R0),
%	retrieve_all(Terms,ST),
%	cputime(R1),
%	R2 is R1 - R0,
%	format("Time used for file ~w : retrieving  : ~f seconds~n",[File,R2]),
%	cputime(X0),
%	all_retrievable(Terms,ST),
%	cputime(X1),
%	X2 is X1 - X0,
%	format("Time used for file ~w : retrievable : ~f seconds~n",[File,X2]),
	cputime(C0),
	count_all(Terms,ST,0,Number),
	cputime(C1),
	C2 is C1 - C0,
	format("Time used for file ~w : count[~d] : ~f seconds~n",[File,Number,C2]),
%	cputime(SC0),
%	simple_count_all(Terms,Terms,0,SNumber),
%	cputime(SC1),
%	SC2 is SC1 - SC0,
%	format("Time used for file ~w : simplecount[~d] : ~f seconds~n",[File,SNumber,SC2]),
	cputime(D0),
	delete_all(Terms,ST,Empty),
	write(Empty),nl,
	cputime(D1),
	D2 is D1 - D0,
	format("Time used for file ~w : delete : ~f seconds~n",[File,D2]),
	true.
	

time_file_insert(File) :-
	cputime(Read0),
	read_terms_file(File,Terms),
	cputime(Read1),
	Read2 is Read1 - Read0,
	format("Time used for file ~w : reading     : ~f seconds~n",[File,Read2]),
	cputime(T),
	insert_all(Terms,_ST),
	cputime(TT),
	Time is TT - T,
	format("Time used for file ~w : inserting   : ~f seconds~n",[File,Time]).

count_file(File) :-
	read_terms_file(File,Terms),
	count_terms(Terms).

count_terms(Terms) :-
	insert_all(Terms,ST),
	count_all(Terms,ST,0,Number),
	format("Magic Number : ~d",[Number]).

simple_count_all([],_,Number,Number).
simple_count_all([Term|Terms],List,Number0,Number2) :-
	copy_term(Term,Copy),
	findall(Term,unif_member(Copy,List),L),
	length(L,Add),
	write('c'),ttyflush,
	Number1 is Number0 + Add,
	simple_count_all(Terms,List,Number1,Number2).

unif_member(Term,[Term0|_]) :-
	sound_unify(Term,Term0).
unif_member(Term,[_|List]) :-
	unif_member(Term,List).


count_all([],_,Number,Number).
count_all([Term|Terms],ST,Number0,Number2) :-
	findall(Term,st_retrieve(Term,ST,_),L),
%	write(Term),nl,
%	write('  '),write(L),nl,
	length(L,Add),
	write('c'),ttyflush,
	Number1 is Number0 + Add,
	count_all(Terms,ST,Number1,Number2).

/*	
:- dynamic blablabla/1.

count_unifiers(Term,ST,Add) :-
	retractall(blablabla(_)),
	asserta(blablabla(0)),
	count_unifiers(Term,ST),
	blablabla(Add).

count_unifiers(Term,ST) :-
	st_unify(Term,ST),
	add_u_counter,
	fail.
count_unifiers(_,_).

add_u_counter :-
	retract(blablabla(X)),
	Y is X + 1,
	asserta(blablabla(Y)).

*/

test_file(File) :-
	read_terms_file(File,Terms),
	insert_all(Terms,ST),
	retrieve_all(Terms,ST),
	all_retrievable(Terms,ST).

test_file_del(File) :-
	read_terms_file(File,Terms),
	insert_all(Terms,ST),
	test_delete(Terms,ST,DST),
	write(DST).

delete_all([],ST,ST).
delete_all([Term|Terms],ST0,ST2) :-
	write('d'),ttyflush,
	st_delete_term(Term,ST0,ST1,identical),
	delete_all(Terms,ST1,ST2).

test_delete([],ST,ST).
test_delete([Term|Terms],ST0,ST2) :-
	write('d'),ttyflush,
	st_delete_term(Term,ST0,ST1),
	check_delete(Term,ST0,ST1),
	test_delete(Terms,ST1,ST2).

check_delete(Term,ST0,ST1) :-
	( (st_unify(T,ST1),Term == T)->
	    write('Term not deleted'),nl,
	    fail
	;
	    ((st_unify(T,ST0),Term == T) ->
		fail
	    ;
		write('Term not in set'),nl,
		fail
	    )
	;
	    fail
	).
check_delete(_,_,_).

test_print_ins(File) :-
	read_terms_file(File,Terms,RT),
	insert_all(Terms,ST),
	print_subst_tree(ST,RT).


all_retrievable([],_ST).
all_retrievable([T|Terms],ST) :-
	st_unifiable(T,ST),
	all_retrievable(Terms,ST).

st_unifiable(T,ST) :-
	(st_unify(T,ST) ->
	    write('r'),ttyflush,
	    true
	;
	    format("Damned ~w not retrievable ~n",[T])
	),
	fail.
st_unifiable(_T,_ST).


retrieve_all(Terms,ST) :-
	st_unify(T,ST),
	was_inserted(Terms,T),
	fail.
retrieve_all(_Terms,_ST).

was_inserted(Terms,T) :-
	is_member(Terms,T),
	write('v'),ttyflush,!.
was_inserted(_Terms,T) :-
	format("Damned ~w not found~n",[T]).

is_member([T0|Terms],T) :-
	(T0 == T ->
	    true
	;
	    is_member(Terms,T)
	).

% io-stuff
read_terms_file(File,Terms) :-
	read_terms_file(File,Terms,_).

read_terms_file(File,Terms1,RL) :-
	open(File,read,Stream),
	read_terms_stream(Stream,Terms0),
	convert_terms(Terms0,Terms1,RL),
	close(Stream).

read_terms_stream(Stream,Terms) :-
	read(Stream,Term),
	(Term \== end_of_file ->
	    [Term|TermsRest] = Terms,
	    read_terms_stream(Stream,TermsRest)
	;
	    Terms = []
	).


convert_terms(Terms0,Terms1,RL) :-
	get_variables(Terms0,[],AtomsVars),
	make_replace_list(AtomsVars,RL),
	replace_terms(Terms0,RL,Terms1).

replace_terms([],_,[]).
replace_terms([T|Ts],RL,[RT|RTs]) :-
	replace_term(T,RL,RT),
	replace_terms(Ts,RL,RTs).

replace_term(T,_RL,RT) :-
	var(T),!, % shouldn't happen
	T = RT.
replace_term(T,RL,RT) :-
	atomic(T),!,
	(member(T-V,RL) ->
	    RT = V
	;
	    RT = T
	).
replace_term(T,RL,RT) :-
	functor(T,F,A),
	functor(RT,F,A),
	replace_term(A,T,RL,RT).

replace_term(0,_,_,_) :- !.
replace_term(A,T,RL,RT) :-
	arg(A,T,ST),
	arg(A,RT,SRT),
	replace_term(ST,RL,SRT),
	A1 is A-1,
	replace_term(A1,T,RL,RT).




make_replace_list([],[]).
make_replace_list([AV|AVs],[AV-_|RL]) :-
	make_replace_list(AVs,RL).

get_variables([],AV,AV).
get_variables([Term|Terms],AV0,AV2) :-
	get_variables_term(Term,AV0,AV1),
	get_variables(Terms,AV1,AV2).
	

get_variables_term(Term,AV0,AV1) :-
	var(Term),!, % shouldn't happen, but anyway
	AV0 = AV1.
get_variables_term(Term,AV0,AV1) :-
	atomic(Term),!,
	(is_variable(Term) ->
	    (member(Term,AV0) ->
		AV0 = AV1
	    ;
		AV1 = [Term|AV0]
	    )
	;
	    AV0 = AV1
	).
get_variables_term(Term,AV0,AV1) :-
	functor(Term,_,A),
	get_variables_term(A,Term,AV0,AV1).

get_variables_term(0,_Term,AV,AV) :- !.
get_variables_term(A,Term,AV0,AV2) :-
	arg(A,Term,SubTerm),
	get_variables_term(SubTerm,AV0,AV1),
	A1 is A-1,
	get_variables_term(A1,Term,AV1,AV2).
	
is_variable(Atom) :-
	name(Atom,[Indicator|_]),
	is_variable_char(Indicator).

is_variable_char(0'u).
is_variable_char(0'v).
is_variable_char(0'w).
is_variable_char(0'x).
is_variable_char(0'y).
is_variable_char(0'z).
