%---------------------------------------------------------------------------
% modSubsumes(+CLAUSE1,+CLAUSE2)
% test if CLAUSE1 subsumes CLAUSE2
% 
% Author:     Ullrich Hustadt

modSubsumes(cl(Literals1,char(PT1,_VL1,CRL1,FL1,_W1)),
	    cl(Literals2,char(PT2,_VL2,CRL2,FL2,_W2)),
	    Mode,Clause) :-
	ord_subset(CRL1,CRL2),
	ord_subset(FL1,FL2),
	!,
	(matchLiterals(Literals1,[],subsumption,Literals2,[],
  	               Mode,ILL1,ILL2) ->
	    true
	;
	    matchLiterals(Literals1,[],subsumCheck,Literals2,[],
	                  Mode,ILL1,ILL2)
	),
	!,
        subsumptionResult(Mode,ILL1,ILL2,PT1,PT2,Clause).
modSubsumes(_W1-C1,cl(L2,C2),M,C) :-
        modSubsumes(C1,cl(L2,C2),M,C).
modSubsumes(cl(L1,C1),_W2-C2,M,C) :-
	modSubsumes(cl(L1,C1),C2,M,C).
modSubsumes(_W1-C1,_W2-C2,M,C) :-
	modSubsumes(C1,C2,M,C).


testSubsumes(R) :-
	modSubsumes(cl([not(grandparent(_915)),concept17(_915)],char([c01],[_915],[grandparent,concept17],[],3)),
	cl([not(grandparent(f1(_162237))),concept16(_162237),concept17(_162237)],char([c00],[_162237],[grandparent,concept16,concept17],[f1],4)),
	subsumCheck,R).

subsumptionResult(subsumCheck,_ILL1,_ILL2,_PT1,_PT2,
	          cl([],char(1,2,3,4,5))).
subsumptionResult(subsumption,_ILL1,_ILL2,_PT1,_PT2,
	          cl([],char(1,2,3,4,5))).
subsumptionResult(subsumResolv,ILL1,ILL2,PT1,PT2,
                  N-Clause) :-
	append(ILL1,ILL2,Literals),
%	append(PT1,PT2,PT),
	constructProofTerm(PT1,PT2,subRes,PT),
	computeCharacteristic(cl(Literals),PT,Clause),
	Clause = cl(_,char(_,_,_,_,N)).
	                      
createVarList(0,[]) :-
	!.
createVarList(N,[_X|L]) :-
	M is N-1,
	createVarList(M,L).

%nth(0,[X|_L],X) :-
%	!.
%nth(N,[_X|L],Y) :-
%	M is N-1,
%	nth(M,L,Y).

matchLiterals(       [GL|GLL], GLLDone,
              Mode1, ILL,      ILLDone1,
	      Mode3, ILLRest3, ILLDone3) :-
	matchLiteral(GL,GLLDone,Mode1,ILL,ILLDone1,
	             MatchedLiteral,ILLRest,Mode2),
        append(MatchedLiteral,ILLDone1,ILLDone2),
	matchLiterals(GLL,[GL|GLLDone],Mode2,ILLRest,ILLDone2,
	              Mode3,ILLRest3,ILLDone3).
matchLiterals(      [],  _,
              Mode, ILL, ILLDone,
	      Mode, ILL, ILLDone).

matchLiteral(GL,GLLDone, Mode, [IL|ILL], ILLDone,
             [IL],ILL,Mode) :-
	subsumes_chk(h(GL,GLLDone),h(IL,ILLDone)).
matchLiteral(GL,GLLDone,subsumCheck,[IL|ILL],ILLDone,
	     [],ILL,subsumResolv) :-
	modNegateLiterals([GL],[NGL]),
	subsumes_chk(h(NGL,GLLDone),h(IL,ILLDone)).
matchLiteral(GL,GLLDone,Mode1,[IL|ILL],ILLDone,
             MatchedLiteral,[IL|ILLRest],Mode2) :-
	matchLiteral(GL,GLLDone,Mode1,ILL,ILLDone,
                     MatchedLiteral,ILLRest,Mode2).

%---------------------------------------------------------------------------
% newStillman(+CLAUSE1,+CLAUSE2)
% true if CLAUSE1 subsumes CLAUSE2.
% The algorithm has first been proposed by Stillman
%
% Author: Patrick Brandmeier
% 
% 
%
% newStillman([],_ArgD) :- !.
% newStillman([LitC|R],ArgD) :-
%	newStillman1(LitC,ArgD),
%	newStillman(R,ArgD).
%
% newStillman1(LitC,[LitD|R]) :-
%	LitC = LitD.
% newStillman1(LitC,[_|R]) :-
%	newStillman1(LitC,R).

%---------------------------------------------------------------------------
% groundingVarList(+-VARLIST,+NUMBER)
% replaces the variables in VARLIST by unique constants.
%
% Author:     Ullrich Hustadt

newVarList([],[],_FreeVars).
newVarList([in(M,C,A1)|L1],[in(M,C,A2)|L2],FreeVars) :-
	newVarTerm(A1,A2,FreeVars),
	newVarList(L1,L2,FreeVars).
newVarList([in(M,C,A1,B1)|L1],[in(M,C,A2,B2)|L2],FreeVars) :-
	newVarTerm(A1,A2,FreeVars),
	newVarTerm(B1,B2,FreeVars),
	newVarList(L1,L2,FreeVars).
newVarList([not(Lit1)|L1],[not(Lit2)|L2],FreeVars) :-
	newVarList([Lit1|L1],[Lit2|L2],FreeVars).

newVarTerm(T,V,FreeVars) :-
	number(T),
	nth(T,FreeVars,V).
newVarTerm(T,T,_) :-
	atomic(T).
newVarTerm(T1,T2,FreeVars) :-
	T1 =.. [F|TL1],
	newVarTerms(TL1,TL2,FreeVars),
	T2 =.. [F|TL2].

newVarTerms([],[],_FreeVars).
newVarTerms([T1|TL1],[T2|TL2],FreeVars) :-
	newVarTerm(T1,T2,FreeVars),
	newVarTerms(TL1,TL2,FreeVars).
   

groundingVarList([],N,N).
groundingVarList([V|VL],N,M) :-
	atomic(V),
	!,
	groundingVarList(VL,N,M).
groundingVarList([V|VL],N,M) :-
	var(V),
	genconstant(N,V,M),
	groundingVarList(VL,M,_K).

genconstant(M,V,N) :-
	N is M+1,
	V = M.
%	name(M,N1),
%	name(V,[0's|N1]).
